@@ -1,8 +1,13 @@
-/* Buttons To Press btp-6.4.0;  Enrique Vidal, Martin Samuelson, Sep-2002).
-   - Move all LIRC-dependent code to rc.cxx
-   - Improved multiple layouts or "views" using Fl_Group.
-   - Correct auto-scaling with the new multiple layouts code
-   - Improved beeping code.
+/* Buttons To Press btp-6.4.1;  Enrique Vidal, March-2003.
+   - New code for LED flashing (only works with 8.1+ kernels),
+   - A new keyword "copyreset" to specify a "copy-without-sending" default,
+   - All the string malloc code now in a generic function: getStrMem.
+
+   Buttons To Press btp-6.4.0;  Enrique Vidal, Martin Samuelson, Sep-2002).
+   - Move all LIRC-dependent code to rc.cxx,
+   - Improved multiple layouts or "views" using Fl_Group,
+   - Correct auto-scaling with the new multiple layouts code,
+   - Improved beeping code,
    - Fixed a few memory leaks.
 
    Buttons To Press btp-6.3.3;  Enrique Vidal, evidal@iti.upv.es, Apr, 2002.
@@ -107,14 +112,17 @@
 Display *Xdisp;                // For changing the Shift_R code
 
 
-/*---------------------------- prototype of the .ini file loading function */
-int loadLayoutFile(char *fname);
+/*----------------------------- forward declarations (function prototypes) */
+int  loadLayoutFile(char *fname);
+void cb_resetCopyBlink();
+char *getStrMem(char *buff, int bytes);
 
 /*------------------------------------------------- For the sending rutine */
 // extern int rcsend(char *command, int fd);              // now in "rc.h"
 // int fd; // The socket.                                // Now in rc.cxx
 
 /*------------------------------------------------- Global vars and tables */
+FILE *consoleFile;             // For the LEDs
 Display *XdispBeep;            // For the beeps
 
 #define maxIRs 16              // Maximum number of remote layouts (.ini files)
@@ -142,11 +150,14 @@ int show_toolbar = 1;          // show_toolbar flag
 #define maxHDbutt 6            // Number of available hardbuttons
 char *hdcomm[maxHDbutt];       // The hadrbutton commands
 
+char *copyReset=NULL;          // The copyreset command
+char *copyResetLabel=NULL;     // The copyreset label
 
 /*-------------------------------------------------------------------------*/
 /*-------------------------------------- Toolbar curr/next button handling */
 void labelNextCurr() {
-  int j;  char *id1=(char *)malloc(5), *id2=(char *)malloc(5);
+  int j;  
+  char *id1=NULL, *id2=NULL;  id1 = getStrMem(id1,5); id2 = getStrMem(id2,5);
 
   if (! show_toolbar) return;
   id1[0]='\0'; strncat(id1, layoutID[activeIR],4);
@@ -170,11 +181,11 @@ void cb_next(Fl_Widget *widget, void *data) {
   b->label("Copy"); b->clear(); b->show(); b->redraw();
   prevData=NULL; wantCopy=0;
 
-  loadLayoutFile(fname[activeIR]);  labelNextCurr();
+  loadLayoutFile(fname[activeIR]);  labelNextCurr();  cb_resetCopyBlink();
 }
 
 void cb_layout(Fl_Widget *w, void *data) {
-int i; Fl_Button *b;
+  int i; Fl_Button *b;
 
   i=0; while (i<numIRs && strcmp(layoutID[i],(char *)data)) i++; 
   if (i < numIRs) {
@@ -182,14 +193,14 @@ int i; Fl_Button *b;
     b=CopyBlink;  b->label("Copy"); b->clear(); b->show(); b->redraw();
     prevData=NULL; wantCopy=0;
 
-    loadLayoutFile(fname[activeIR]);  labelNextCurr();
+    loadLayoutFile(fname[activeIR]);  labelNextCurr();  cb_resetCopyBlink();
   }
 }
 
 /*------------------------------------------------------------------- Beeps */
                     /* From beep.c, (C) 2001 James Fenn <jfenn@uklinux.net> */
 void beep(int pitch, int duration, int permLevel) {
-//Display *XdispBeep;                            // Global; assumedly opon.
+//Display *XdispBeep;                            // Global; assumedly open.
   XKeyboardControl vals;
 
   if (wantBeeps < permLevel) return;
@@ -214,8 +225,8 @@ void cb_resetCopyBlink() {    // Timeout to reset CopyBlink toolbar button
   // Remove timeout, just in case we are called from curr/next button cb's
   Fl::remove_timeout((Fl_Timeout_Handler)cb_resetCopyBlink); 
   Fl_Button *b=CopyBlink;
-  b->label("Copy"); b->clear(); b->show(); b->redraw();
-  prevData=NULL; wantCopy=0;
+  b->label(copyResetLabel); b->clear(); b->show(); b->redraw(); 
+  prevData=copyReset;  wantCopy=0;
 //fprintf(stderr,"--> Timeout exhausted!\n"); 
 }
 void CopyBlinkTimeout() {
@@ -226,20 +237,33 @@ void CopyBlinkTimeout() {
 
 void cb_copy(Fl_Widget *widget, void *data) {
   if (show_toolbar) {
-    Fl_Button *b=CopyBlink;
-    b->label("Copy"); b->set(); b->show(); b->redraw();
+     Fl_Button *b=CopyBlink;
+     b->label("Copy"); b->set(); b->show(); b->redraw();
   }
   prevData=NULL; wantCopy=1; CopyBlinkTimeout();
 }
 
+void ledFlash(int autoStop, int offPeriod, int onPeriod, int blinkNum) {
+#ifdef AGVR3
+  if (consoleFile == NULL) return;
+  fprintf(consoleFile, "\033[16;%d]\033[17;%d]\033[19;%d]\033[18;%d]\n", 
+                               onPeriod, offPeriod, blinkNum, autoStop);
+#endif
+}
+void ledFlashOff() {
+#ifdef AGVR3
+  if (consoleFile == NULL) return; fprintf(consoleFile, "\033[18;0]\n");
+#endif
+}
+
 void blinker(int OnOff, Fl_Button *b) {
-static char *label;
+  static char *label;
 //fprintf(stderr,"--> %s\n", b->label()); 
   if (OnOff) {
-        label = strdup(b->label());
-        beep(1000, 60, 1); b->label("))))"); b->set();
+     ledFlash(1,3,1,0);  label = strdup(b->label());
+     beep(1200, 60, 1);  b->label("))))"); b->set();
   }
-  else {b->label(label); b->clear(); beep(1000, 200, 1);}
+  else {b->label(label);  b->clear();  beep(1200, 200, 1);  ledFlashOff();}
   b->redraw(); Fl::flush();
 }
 
@@ -254,14 +278,15 @@ void cb_button(Fl_Widget *w, void *data) {
     int isLong = (strlen((char *)data) > maxLength);        // macro timing!!
     if (isLong) blinker(1, b);
     rcsend((char *)data);
+    ledFlash(3,0,1,1);
 //  beep(1000, 8, 2);   // all buttons would beep
     if (isLong) {blinker(0, b); usleep(600000);} 
     else if (show_toolbar) c->clear();
   }
 //fprintf(stderr,"--->   %s\n", b->label()); 
   if (show_toolbar) {
-    c->label(b->label());  // Change Copy-label to remind what's been copied.
-    if (w != c) c->set(); c->redraw(); Fl::flush();
+     c->label(b->label());  // Change Copy-label to remind what's been copied.
+     if (w != c) c->set(); c->redraw(); Fl::flush();
   }
   CopyBlinkTimeout();
 }
@@ -274,7 +299,7 @@ void rcHdSend(char *data) {
      if (prevData != NULL)
         cb_button(CopyBlink, prevData); beep(300,60,1);
   }
-  else rcsend(data);
+  else { rcsend(data); ledFlash(3,0,1,1); }
 }
 
 /*------------------------------------------------------- X Error Handling */
@@ -292,6 +317,8 @@ KeySym ks;
   Fl_Window* window = reinterpret_cast<Fl_Window*>(data);
   window->hide();
   XCloseDisplay(XdispBeep);
+  fclose(consoleFile);
+
 #ifdef AGVR3
   ks = XK_Shift_L;
 #else
@@ -314,10 +341,18 @@ void cb_timeout() {
   if (i<numIRs) Fl::add_timeout(iniLoadTime, (Fl_Timeout_Handler)cb_timeout);
 }
 
+/*------------------------- Get memory for a string in a controlled manner */
+char *getStrMem(char *buff, int bytes) {
+  if (buff != NULL) free(buff);
+  if ((buff = (char *)malloc(bytes)) == NULL) 
+     {btpStderr("Out of memory: %d\n", bytes); exit(EXIT_FAILURE);}
+  return(buff);
+}
+
 /*---------------------------------------------------- Define a new Layout */
 int loadLayoutFile(char *fname) {
 FILE *rcfile;
-char *command, layout_id[64];
+char layout_id[64];
 char t[1024];                                   // The .init file read buffer
 int labelsize = 10;
 int layoutgood = 0;
@@ -327,6 +362,10 @@ Fl_Widget *b;
 static int first=1;
 static Fl_Group *btpgroup;
 
+  if (copyReset != NULL) free(copyReset);  copyReset=NULL;
+  if (copyResetLabel != NULL) free(copyResetLabel);
+  copyResetLabel = strdup("Copy");  prevData=NULL;
+
   //fprintf(stderr,"llf: %s\n", fname);
 
   //if (btpgroup != NULL) btpgroup->~Fl_Group();
@@ -374,6 +413,22 @@ static Fl_Group *btpgroup;
         labelsize = atoi(strtok(parseable, ",\n"));
       }
       // --------------------------------------------------------------------
+      if (!strncasecmp(t, "copyreset", 9)) { // Parse copyreset specification
+        char *parseable = index(t, '=') + 1;
+        char *label = strdup(strtok(parseable, ","));
+        char *remote = strdup(strtok(NULL, ","));
+        char *remoteCode = strdup(strtok(NULL, "\n"));
+
+        bytes = strlen(remote)+strlen(remoteCode)+3;
+        copyReset = getStrMem(copyReset, bytes);
+        sprintf(copyReset, "%s %s", remote, remoteCode);
+
+        if (copyResetLabel != NULL) free(copyResetLabel);
+        copyResetLabel = strdup(label);
+        free(remote); free(remoteCode); free(label);
+        //fprintf(stderr, "--- - %s - %s -\n", copyResetLabel, copyReset);
+      }
+      // --------------------------------------------------------------------
       if (!strncasecmp(t, "hardbutton", 10)) {           // Parse hardbuttons
         int j=UnknownInd;
         char *parseable = index(t, '=') + 1;
@@ -382,7 +437,7 @@ static Fl_Group *btpgroup;
         char *remoteCode;
         if (!strncasecmp(remote, ".prev", 5)) remoteCode = strdup("");
         else remoteCode = strdup(strtok(NULL, "\n"));
-  
+
         if      (strcmp(key, Shift_L)   == 0) j=Shift_LInd;
         else if (strcmp(key, Shift_R)   == 0) j=Shift_RInd;
         else if (strcmp(key, Left)      == 0) j=LeftInd;
@@ -391,11 +446,9 @@ static Fl_Group *btpgroup;
         else if (strcmp(key, Page_Down) == 0) j=Page_DownInd; 
   
         if (j>=0 && j<maxHDbutt) {
-          bytes = strlen(remote)+strlen(remoteCode)+3;
-          if (hdcomm[j] != NULL) free(hdcomm[j]);
-          if ((hdcomm[j] = (char *)malloc(bytes)) == NULL) 
-            {btpStderr("Out of memory: %d\n", bytes); exit(EXIT_FAILURE);}
-          sprintf(hdcomm[j], "%s %s", remote, remoteCode);
+           bytes = strlen(remote)+strlen(remoteCode)+3;
+           hdcomm[j] = getStrMem(hdcomm[j], bytes);
+           sprintf(hdcomm[j], "%s %s", remote, remoteCode);
         }
         else btpStderr("Hard button unknown: %s\n", key);
         //fprintf(stderr, "--- %d %s - %s-\n", j, key, hdcomm[j]);
@@ -414,8 +467,7 @@ static Fl_Group *btpgroup;
         char *remoteCode = strdup(strtok(NULL, "\n"));
 
         bytes = strlen(remote)+strlen(remoteCode)+3;
-        if ((command = (char *)malloc(bytes)) == NULL)
-           {btpStderr( "Out of memory: %d\n", bytes); exit(EXIT_FAILURE);}
+        char *command; command = getStrMem(NULL, bytes);
         sprintf(command, "%s %s", remote, remoteCode);
         
 	if(!first) {
@@ -428,7 +480,7 @@ static Fl_Group *btpgroup;
         b->align(FL_ALIGN_WRAP | FL_ALIGN_CLIP);
         b->labelsize(labelsize);
         if (!strncasecmp(remote, "layout", 6))          // Call layout button
-	     {b->callback(cb_layout, remoteCode); free(command);}
+	     {b->callback(cb_layout, remoteCode); free(command); command=NULL;}
         else {b->callback(cb_button, command); free(remoteCode);}
         free(remote);
         if (!first) btpgroup->add(b); b->redraw(); // b->show(); 
@@ -465,28 +517,27 @@ int i=0, j;
     
     for (i=j=0; ((i<numIRs) && (i<maxIRs)); i++)    {     FILE *rcfile; 
       if (argc>1) fname[j] = argv[i+1];
-//    fprintf(stderr,"%d: %s  %s\n", i+1, argv[i+1], fname[i]);
+      //fprintf(stderr,"%d: %s  %s\n", i+1, argv[i+1], fname[i]);
       
       if ((rcfile = fopen(fname[j], "r"))) { char t[64];    // Set layout IDs
-        if (layoutID[j] != NULL) free(layoutID[j]);
-        if ((layoutID[j] = (char *)malloc(2)) == NULL)
-           {btpStderr( "Out of memory: %d\n", 2); exit(EXIT_FAILURE);}
-        sprintf(layoutID[j], "%d", j);                          // Default ID
-        fgets(t, 64, rcfile);
-        if (!strncasecmp(t, "layout", 6)) {               // Parse and set ID
-          char *parseable = index(t, '=') + 1;
-          layoutID[j] = strdup(strtok(parseable, ",\n"));
-        }
-        j++;
+         layoutID[j] = getStrMem(layoutID[j], 2);
+         sprintf(layoutID[j], "%d", j);                         // Default ID
+
+         fgets(t, 64, rcfile);
+         if (!strncasecmp(t, "layout", 6)) {              // Parse and set ID
+            char *parseable = index(t, '=') + 1;
+            layoutID[j] = strdup(strtok(parseable, ",\n"));
+         }
+         j++;
       }
-      else btpStderr("Could not open %s\n", fname[j]);  // Mild error...
-//    fprintf(stderr,"%d,%d: %s  %s\n", i,j-1, fname[j-1], layoutID[j-1]);
-    } 
+      else btpStderr("Could not open %s\n", fname[j]);       // Mild error...
+      //fprintf(stderr,"%d,%d: %s  %s\n", i,j-1, fname[j-1], layoutID[j-1]);
+    }
     i=0; numIRs=j;
 
     // btpwindow = new Fl_Double_Window(0,0,160,240,"BTP 6.3.3");
     // Double_Window is much better, but not seems supported by WidgetFactory
-    btpwindow = WidgetFactory::new_window("BTP 6.4.0");
+    btpwindow = WidgetFactory::new_window("BTP 6.4.1");
     layoutgood = loadLayoutFile(fname[0]);
     if (show_toolbar) {
        Fl_Dockable_Window* toolbar = WidgetFactory::new_toolbar();
@@ -497,7 +548,7 @@ int i=0, j;
        curr=Widget_Factory::new_button("",             cb_curr, btpwindow);
        next=Widget_Factory::new_button("",             cb_next, btpwindow);
        btpwindow->add_dockable(toolbar, 1);
-       toolbar->end(); 
+       toolbar->end();
     }
     btpwindow->end(); btpwindow->show();
 
@@ -513,6 +564,10 @@ int i=0, j;
     XChangeKeyboardMapping(Xdisp, Shift_R_Code, 1, &ks, 1);
     XCloseDisplay(Xdisp);
     XdispBeep=XOpenDisplay(NULL);   // For the beeps; to be closed by cb_done
+    consoleFile = fopen("/dev/console", "w");             // for LED flashing
+    if (consoleFile == NULL) {                                           
+       btpStderr("Error opening /dev/console: %s.\n", strerror(errno));
+    }
 
     Fl::add_timeout(iniLoadTime, (Fl_Timeout_Handler)cb_timeout);
 
